(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('log4js'), require('node:fs/promises'), require('node:path'), require('node:fs'), require('fast-xml-parser'), require('node:crypto')) :
    typeof define === 'function' && define.amd ? define(['exports', 'log4js', 'node:fs/promises', 'node:path', 'node:fs', 'fast-xml-parser', 'node:crypto'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.jsTemplate = {}, global.log4js, global.fs, global.path, global.fs$1, global.fastXmlParser, global.node_crypto));
})(this, (function (exports, log4js, fs, path, fs$1, fastXmlParser, node_crypto) { 'use strict';

    function _interopNamespaceDefault(e) {
        var n = Object.create(null);
        if (e) {
            Object.keys(e).forEach(function (k) {
                if (k !== 'default') {
                    var d = Object.getOwnPropertyDescriptor(e, k);
                    Object.defineProperty(n, k, d.get ? d : {
                        enumerable: true,
                        get: function () { return e[k]; }
                    });
                }
            });
        }
        n.default = e;
        return Object.freeze(n);
    }

    var log4js__namespace = /*#__PURE__*/_interopNamespaceDefault(log4js);
    var fs__namespace = /*#__PURE__*/_interopNamespaceDefault(fs);
    var path__namespace = /*#__PURE__*/_interopNamespaceDefault(path);
    var fs__namespace$1 = /*#__PURE__*/_interopNamespaceDefault(fs$1);

    /******************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */
    /* global Reflect, Promise, SuppressedError, Symbol */


    function __awaiter(thisArg, _arguments, P, generator) {
        function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
        return new (P || (P = Promise))(function (resolve, reject) {
            function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
            function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
            function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
            step((generator = generator.apply(thisArg, _arguments || [])).next());
        });
    }

    typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
        var e = new Error(message);
        return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
    };

    const XML_HEADER = '<?xml version="1.0" encoding="UTF-8"?>';
    const INDENT = "    ";

    const XML_PARSER_OPTION = {
        commentPropName: "!---",
        ignoreAttributes: false,
        ignoreNameSpace: false,
        parseTagValue: false,
        parseNodeValue: false,
        parseAttributeValue: false,
        trimValues: true,
        processEntities: false,
        cdataPropName: "![CDATA[",
    };
    const JSON_PARSER_OPTION = Object.assign(Object.assign({}, XML_PARSER_OPTION), { format: true, indentBy: INDENT, suppressBooleanAttributes: false, suppressEmptyNode: false });

    function buildReassembledFile(combinedXmlContents, filePath, xmlElement, xmlNamespace) {
        return __awaiter(this, void 0, void 0, function* () {
            let finalXmlContent = combinedXmlContents.join("\n");
            finalXmlContent = finalXmlContent.replace(/<\?xml version="1.0" encoding="UTF-8"\?>/g, "");
            finalXmlContent = finalXmlContent.replace(new RegExp(`<${xmlElement}[^>]*${xmlNamespace ? `( xmlns="${xmlNamespace}")?` : ""}>`, "g"), "");
            finalXmlContent = finalXmlContent.replace(new RegExp(`</${xmlElement}>`, "g"), "");
            finalXmlContent = finalXmlContent.replace(/<!\[CDATA\[\s*([\s\S]*?)\s*]]>/g, (match, cdataContent) => {
                const trimmedContent = cdataContent.trim();
                const lines = trimmedContent.split("\n");
                const indentedLines = lines.map((line) => line.replace(/^\s*/, ""));
                return `<![CDATA[\n${INDENT}${indentedLines.join(`\n${INDENT}`)}\n]]>`;
            });
            finalXmlContent = finalXmlContent.replace(/(\n\s*){2,}/g, `\n${INDENT}`);
            const openTag = xmlNamespace
                ? `<${xmlElement} xmlns="${xmlNamespace}">`
                : `<${xmlElement}>`;
            const closeTag = `</${xmlElement}>`;
            yield fs__namespace.writeFile(filePath, `${XML_HEADER}\n${openTag}${finalXmlContent}${closeTag}`);
            logger.debug(`Created reassembled file: ${filePath}`);
        });
    }

    function buildNestedElements(element, indentLevel = 2) {
        const xmlBuilder = new fastXmlParser.XMLBuilder(JSON_PARSER_OPTION);
        const xmlString = xmlBuilder.build(element);
        const formattedXml = xmlString
            .split("\n")
            .map((line) => `${" ".repeat(indentLevel * INDENT.length)}${line}`)
            .join("\n")
            .trimEnd();
        return formattedXml;
    }

    const xmlParser = new fastXmlParser.XMLParser(XML_PARSER_OPTION);
    class ReassembleXMLFileHandler {
        processFilesInDirectory(dirPath) {
            return __awaiter(this, void 0, void 0, function* () {
                const combinedXmlContents = [];
                let rootResult = undefined;
                const files = yield fs__namespace.readdir(dirPath);
                files.sort((fileA, fileB) => {
                    const fullNameA = fileA.split(".")[0].toLowerCase();
                    const fullNameB = fileB.split(".")[0].toLowerCase();
                    return fullNameA.localeCompare(fullNameB);
                });
                for (const file of files) {
                    const filePath = path__namespace.join(dirPath, file);
                    const fileStat = yield fs__namespace.stat(filePath);
                    if (fileStat.isFile() && filePath.endsWith(".xml")) {
                        const xmlContent = yield fs__namespace.readFile(filePath, "utf-8");
                        const xmlParsed = xmlParser.parse(xmlContent);
                        const rootResultFromFile = yield this.processFilesForRootElement(xmlParsed);
                        if (rootResultFromFile && !rootResult) {
                            rootResult = rootResultFromFile;
                        }
                        const combinedXmlString = buildNestedElements(xmlParsed, 0);
                        combinedXmlContents.push(combinedXmlString);
                    }
                }
                return [combinedXmlContents, rootResult];
            });
        }
        processFilesForRootElement(xmlParsed) {
            return __awaiter(this, void 0, void 0, function* () {
                const rootElementName = Object.keys(xmlParsed)[1];
                const rootElement = xmlParsed[rootElementName];
                let rootElementNamespace;
                if (rootElement["@_xmlns"] !== undefined) {
                    rootElementNamespace = String(rootElement["@_xmlns"]);
                }
                else {
                    rootElementNamespace = undefined;
                }
                if (rootElementName !== undefined && rootElementName.length > 0) {
                    return [rootElementName, rootElementNamespace];
                }
                return undefined;
            });
        }
        reassemble(xmlAttributes) {
            return __awaiter(this, void 0, void 0, function* () {
                const { xmlPath, fileExtension, postPurge = false } = xmlAttributes;
                const combinedXmlContents = [];
                const fileStat = yield fs__namespace.stat(xmlPath);
                if (!fileStat.isDirectory()) {
                    logger.error(`The provided xmlPath ${xmlPath} to reassemble is not a directory.`);
                    return;
                }
                logger.debug(`Parsing directory to reassemble: ${xmlPath}`);
                const filesInxmlPath = yield fs__namespace.readdir(xmlPath);
                filesInxmlPath.sort((fileA, fileB) => {
                    const fullNameA = fileA.split(".")[0].toLowerCase();
                    const fullNameB = fileB.split(".")[0].toLowerCase();
                    return fullNameA.localeCompare(fullNameB);
                });
                let rootResult = undefined;
                for (const file of filesInxmlPath) {
                    const filePath = path__namespace.join(xmlPath, file);
                    const fileStat = yield fs__namespace.stat(filePath);
                    if (fileStat.isFile() && filePath.endsWith(".xml")) {
                        const xmlContent = yield fs__namespace.readFile(filePath, "utf-8");
                        const xmlParsed = xmlParser.parse(xmlContent);
                        rootResult = yield this.processFilesForRootElement(xmlParsed);
                        const combinedXmlString = buildNestedElements(xmlParsed, 0);
                        combinedXmlContents.push(combinedXmlString);
                    }
                    else if (fileStat.isDirectory()) {
                        const [subdirectoryContents, subdirectoryRootResult] = yield this.processFilesInDirectory(filePath);
                        combinedXmlContents.push(...subdirectoryContents);
                        if (subdirectoryRootResult && !rootResult) {
                            rootResult = subdirectoryRootResult;
                        }
                    }
                }
                const parentDirectory = path__namespace.dirname(xmlPath);
                const subdirectoryBasename = path__namespace.basename(xmlPath);
                const fileName = fileExtension
                    ? `${subdirectoryBasename}.${fileExtension}`
                    : `${subdirectoryBasename}.xml`;
                const filePath = path__namespace.join(parentDirectory, fileName);
                if (rootResult !== undefined) {
                    const [rootElementName, rootElementNamespace] = rootResult;
                    yield buildReassembledFile(combinedXmlContents, filePath, rootElementName, rootElementNamespace);
                    if (postPurge)
                        yield fs__namespace.rm(xmlPath, { recursive: true });
                }
                else {
                    logger.error(`A Root Element Name was not found in any files under ${xmlPath}`);
                    return;
                }
            });
        }
    }

    function findUniqueIdElement(element, uniqueIdElements) {
        if (uniqueIdElements === undefined) {
            return getShortHash(element);
        }
        const uniqueIdElementsArray = uniqueIdElements.split(",");
        for (const fieldName of uniqueIdElementsArray) {
            if (element[fieldName] !== undefined) {
                if (typeof element[fieldName] === "string") {
                    return element[fieldName];
                }
            }
        }
        for (const key in element) {
            if (typeof element[key] === "object" && element[key] !== null) {
                const childFieldName = findUniqueIdElement(element[key], uniqueIdElements);
                if (childFieldName !== undefined) {
                    return childFieldName;
                }
            }
        }
        return getShortHash(element);
    }
    function getShortHash(element) {
        const hash = node_crypto.createHash("sha256");
        hash.update(JSON.stringify(element));
        const fullHash = hash.digest("hex");
        return fullHash.slice(0, 8);
    }

    function buildDisassembledFiles(xmlString, metadataPath, uniqueIdElements, baseName, indent, postPurge, parentPath) {
        const xmlParser = new fastXmlParser.XMLParser(XML_PARSER_OPTION);
        const result = xmlParser.parse(xmlString);
        const rootElementName = Object.keys(result)[1];
        if (rootElementName.length === 0) {
            logger.error(`A Root Element Name was not found in ${baseName}.xml`);
            return;
        }
        const rootElement = result[rootElementName];
        let rootElementNamespace;
        if (rootElement["@_xmlns"] !== undefined) {
            rootElementNamespace = String(rootElement["@_xmlns"]);
        }
        else {
            rootElementNamespace = undefined;
        }
        let leafContent = "";
        let leafCount = 0;
        let hasNestedElements = false;
        Object.keys(rootElement)
            .filter((key) => !key.startsWith("@"))
            .forEach((key) => {
            if (Array.isArray(rootElement[key])) {
                for (const element of rootElement[key]) {
                    if (typeof element === "object") {
                        buildNestedFile(element, metadataPath, uniqueIdElements, rootElementName, rootElementNamespace, key, indent);
                        hasNestedElements = true;
                    }
                    else {
                        const fieldValue = element;
                        leafContent += `${indent}<${key}>${String(fieldValue)}</${key}>\n`;
                        leafCount++;
                    }
                }
            }
            else if (typeof rootElement[key] === "object") {
                buildNestedFile(rootElement[key], metadataPath, uniqueIdElements, rootElementName, rootElementNamespace, key, indent);
                hasNestedElements = true;
            }
            else {
                const fieldValue = rootElement[key];
                leafContent += `${indent}<${key}>${String(fieldValue)}</${key}>\n`;
                leafCount++;
            }
        });
        if (!hasNestedElements) {
            logger.error(`The XML file ${baseName}.xml only has leaf elements. This file will not be disassembled.`);
            return;
        }
        if (leafCount > 0) {
            let leafFile = `${XML_HEADER}\n`;
            leafFile += `<${rootElementName}`;
            if (rootElementNamespace) {
                leafFile += ` xmlns="${rootElementNamespace}"`;
            }
            leafFile += `>\n`;
            const sortedLeafContent = leafContent
                .split("\n")
                .filter((line) => line.trim() !== "")
                .sort()
                .join("\n");
            leafFile += sortedLeafContent;
            leafFile += `\n</${rootElementName}>`;
            const leafOutputPath = path__namespace.join(metadataPath, `${baseName}.xml`);
            fs__namespace$1.writeFileSync(leafOutputPath, leafFile);
            logger.debug(`Created disassembled file: ${leafOutputPath}`);
        }
        if (postPurge) {
            const originalFilePath = path__namespace.resolve(`${parentPath}/${baseName}.xml`);
            fs__namespace.unlink(originalFilePath);
        }
    }
    function buildNestedFile(element, metadataPath, uniqueIdElements, rootElementName, rootElementNamespace, parentKey, indent) {
        let elementContent = "";
        const fieldName = findUniqueIdElement(element, uniqueIdElements);
        const outputDirectory = path__namespace.join(metadataPath, parentKey);
        const outputFileName = `${fieldName}.${parentKey}-meta.xml`;
        const outputPath = path__namespace.join(outputDirectory, outputFileName);
        fs__namespace$1.mkdirSync(outputDirectory, { recursive: true });
        elementContent = buildNestedElements(element);
        let decomposeFileContents = `${XML_HEADER}\n`;
        decomposeFileContents += `<${rootElementName}`;
        if (rootElementNamespace) {
            decomposeFileContents += ` xmlns="${rootElementNamespace}"`;
        }
        decomposeFileContents += `>\n`;
        decomposeFileContents += `${indent}<${parentKey}>\n`;
        decomposeFileContents += `${elementContent}\n`;
        decomposeFileContents += `${indent}</${parentKey}>\n`;
        decomposeFileContents += `</${rootElementName}>`;
        fs__namespace$1.writeFileSync(outputPath, decomposeFileContents);
        logger.debug(`Created disassembled file: ${outputPath}`);
    }

    class DisassembleXMLFileHandler {
        disassemble(xmlAttributes) {
            return __awaiter(this, void 0, void 0, function* () {
                const { xmlPath, uniqueIdElements, prePurge = false, postPurge = false, } = xmlAttributes;
                const fileStat = yield fs__namespace.stat(xmlPath);
                if (fileStat.isFile()) {
                    const filePath = path__namespace.resolve(xmlPath);
                    if (!filePath.endsWith(".xml")) {
                        logger.error(`The file path ${filePath} is not an XML file.`);
                        return;
                    }
                    const basePath = path__namespace.dirname(filePath);
                    yield this.processFile({
                        xmlPath: basePath,
                        filePath,
                        uniqueIdElements,
                        prePurge,
                        postPurge,
                    });
                }
                else if (fileStat.isDirectory()) {
                    const files = yield fs__namespace.readdir(xmlPath);
                    for (const file of files) {
                        const filePath = path__namespace.join(xmlPath, file);
                        if (filePath.endsWith(".xml")) {
                            yield this.processFile({
                                xmlPath,
                                filePath,
                                uniqueIdElements,
                                prePurge,
                                postPurge,
                            });
                        }
                    }
                }
            });
        }
        processFile(xmlAttributes) {
            return __awaiter(this, void 0, void 0, function* () {
                const { xmlPath, filePath, uniqueIdElements, prePurge, postPurge } = xmlAttributes;
                logger.debug(`Parsing file to disassemble: ${filePath}`);
                const xmlContent = yield fs__namespace.readFile(filePath, "utf-8");
                const fullName = path__namespace.basename(filePath, path__namespace.extname(filePath));
                const baseName = fullName.split(".")[0];
                let outputPath;
                outputPath = path__namespace.join(xmlPath, baseName);
                if (prePurge) {
                    if (fs__namespace$1.existsSync(outputPath)) {
                        yield fs__namespace.rm(outputPath, { recursive: true });
                    }
                }
                buildDisassembledFiles(xmlContent, outputPath, uniqueIdElements, fullName, INDENT, postPurge, xmlPath);
            });
        }
    }

    function setLogLevel(level) {
        log4js__namespace.getLogger().level = level;
    }
    const logger = log4js__namespace.getLogger();
    log4js__namespace.configure({
        appenders: { disassemble: { type: "file", filename: "disassemble.log" } },
        categories: { default: { appenders: ["disassemble"], level: "error" } },
    });

    exports.DisassembleXMLFileHandler = DisassembleXMLFileHandler;
    exports.ReassembleXMLFileHandler = ReassembleXMLFileHandler;
    exports.logger = logger;
    exports.setLogLevel = setLogLevel;

}));
//# sourceMappingURL=index.umd.js.map
